import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.Writer;
import java.io.IOException;

/**
After each \n the current linenumber will be inserted. The Number will only be printed, if at
least one character follows after it. Extends BufferedWriter for efficiency 
as we are printing each character (even of long strings) by itself.
@see BufferedWriter
@see LineNumberReader
@author Dipl.-Ing. Michael Sonntag
@version 1.0
*/
public class LinenumberWriter extends BufferedWriter
{
    private long currentLinenumber=1;   // Current number of line
    private int increment=1;    // Increment between two consecutive lines
    private boolean afterLF=true;   // If we are after a LF
    private String lineSeparator="\r\n";  // Separartor between two lines
    private int posInSep=0; // Next character in lineSeparator that must be matched
    
/**
Create a character-output stream that prints line numbers at the start of 
each new line.
@param out A Writer
*/
public LinenumberWriter(Writer out)
{
    super(out);
    // Doesn't work: Should return \n, but returns \n\n ?????
//	lineSeparator=(String) java.security.AccessController.doPrivileged(new sun.security.action.GetPropertyAction("line.separator"));
}

/**
Create a character-output stream that prints line numbers at the start of 
each new line. Start of the linenumbers and the increment can be set.
@param out A Writer
@param startNumber The starting number
@param increment The difference between the numbers of two consecutive lines
*/
public LinenumberWriter(Writer out,long startNumber,int increment)
{
    this(out);
    currentLinenumber=startNumber;
    this.increment=increment;
}

/**
Write a single character. If immediately after a LF, prints the number of the 
new line in front of the character.
@exception  IOException  If an I/O error occurs
*/
public void write(int c) throws IOException
{
    // Check first if immediately after a LF -> Write new linenumber
    if(afterLF)
    {
        afterLF=false;  // Must be BEFORE write, else endless recursion!
        write(""+currentLinenumber+": ");
    }
    super.write(c); // Write the character
    // This algorithm will only work, if no character appears twice in lineSeparator!
    // E.g.: Separator=RORX, Text=RORORX; At second O the posInSep will be set to 0, but it should be 2!
    if(c==lineSeparator.charAt(posInSep))
    {   // Check for separator
        posInSep++;
        if(posInSep==lineSeparator.length())
        {
            afterLF=true;
            currentLinenumber+=increment;
            posInSep=0;
        }
    }
    else    // Wrong character, reset to start
        posInSep=0;

}

/**
Write a portion of an array of characters. After each line separator a linenumber will be inserted if at least one character 
follows.

@param  cbuf  A character array
@param  off   Offset from which to start reading characters
@param  len   Number of characters to write

@exception  IOException  If an I/O error occurs
*/
public void write(char[] cbuf,int off,int len) throws IOException
{
    for(int i=off;i<cbuf.length && i<off+len;i++)
        write(cbuf[i]);
}

/**
Write a portion of a String. After each line separator a linenumber will be inserted if at least one character 
follows.

@param  s     String to be written
@param  off   Offset from which to start reading characters
@param  len   Number of characters to be written

@exception  IOException  If an I/O error occurs
*/
public void write(String s,int off,int len) throws IOException
{
    write(s.toCharArray(),off,len);    
}

/**
Returns the number of the current line. After a \n and before any other 
character is written, the number of the line is returned, the next character 
will be printed upon (Even though this linenumber is not yet printed. It will 
be printed when the first character is printed).
@return The number of the current line
*/
public long getCurrentLinenumber()
{
    return currentLinenumber;
}

static public void main(String args[])
{
    LinenumberWriter lw=new LinenumberWriter(new PrintWriter(System.out),10,10);
    try
    {
        lw.write('t');
        lw.write('e');
        lw.write('\n');
        lw.write('s');
        lw.write('\r');
        lw.write('t');
        lw.write('\r');
        lw.write('\n');
        lw.write('a');
        lw.write('\r');
        lw.write('\n');
        lw.write("String ohne LF ");
        lw.write("String mit\r\nLF ");
        lw.write("String mit LF am Ende\r\n");
        lw.write("Current linenumber: "+lw.getCurrentLinenumber()+"\r\n");
        lw.flush(); // Don't close, this would end the program!
        lw=new LinenumberWriter(new PrintWriter(System.out));
        lw.write("First line\r\nSecond line\r\nThird line\r\nFourth line\r\n");
        lw.close();
    }
    catch(IOException e)
    {
        System.err.println(e.toString());
    }
}


}

